local name, addon = ...;
local L = addon.L
local GetCompanionInfo, GetNumCompanions = GetCompanionInfo, GetNumCompanions

addon.MOUNTTYPE_LAND, addon.MOUNTTYPE_FLYING, addon.MOUNTTYPE_AQUATIC = 0, 1, 2
local MOUNTTYPE_LAND = addon.MOUNTTYPE_LAND
local MOUNTTYPE_FLYING = addon.MOUNTTYPE_FLYING
local MOUNTTYPE_AQUATIC = addon.MOUNTTYPE_AQUATIC

local tip = CreateFrame("GameTooltip")
local linecache = {}
tip:SetOwner(WorldFrame, "ANCHOR_NONE")
for i=1,10 do
	linecache[i] = tip:CreateFontString()
	tip:AddFontStrings(linecache[i], tip:CreateFontString())
end

local EPIC, LEGENDARY = L["very fast"], L["extremely fast"]
local RED_LINE = "|cffff2020"


--====================
-- UTILITY FUNCTIONS
--====================
-- Search the player's skills for one named to match the supplied (localized) string
local function GetSkillLine(skill)
	for i=1, GetNumSkillLines() do
		local name, header, expanded, rank = GetSkillLineInfo(i)
		if header and not expanded then
			ExpandSkillHeader(i) -- Pop these open just in case; not worth reclosing
		elseif name == skill then
			return rank
		end
	end
	return 0
end

-- Table of the achievements connected to 310% mounts.
local LegendaryMounts = {
	[885] = false,  -- Ashes of Al'ar
	[886] = false,  -- Season 1 Nether Drake
	[887] = false,  -- Season 2 Nether Drake
	[888] = false,  -- Season 3 Nether Drake
	[2316] = false, -- Season 4 Nether Drake
	[3096] = false, -- Season 5 Frostwyrm
	[3756] = false, -- Season 6 Frostwyrm
	[3757] = false, -- Season 7 Frostwyrm
	[4600] = false, -- Season 8 Frostwyrm
	[2137] = false, -- Plagued Proto-Drake (Glory of the Raider 10)
	[2138] = false, -- Black Proto-Drake (Glory of the Raider 25)
	[2957] = false, -- Rusted Proto-Drake (Glory of Ulduar 10)
	[2958] = false, -- Ironbound Proto-Drake (Glory of Ulduar 25)
	[4602] = false, -- Bloodbathed Frostbrood Vanquisher (Glory of Icecrown 10)
	[4603] = false, -- Icebound Frostbrood Vanquisher (Glory of Icecrown 25)
	[2144] = false, -- Violet Proto-Drake (Alliance)
	[2145] = false, -- Violet Proto-Drake (Horde)
}

-- Check each of the achievements in the above table. If the player has one, mark it to make later lookups faster.
local function Has310Mount()
	for k,v in pairs(LegendaryMounts) do
		if v == true or select(4,GetAchievementInfo(k)) == true then
			LegendaryMounts[k] = true
			return true
		end
	end
	return false
end

local function SetTooltip(spellID)
	tip:ClearLines()
	if not tip:IsOwned(WorldFrame) then
		tip:SetOwner(WorldFrame, "ANCHOR_NONE")
	end
	tip:SetHyperlink(GetSpellLink(spellID))
end
--====================
-- UPDATE FUNCTIONS
--====================
local function USABLE_ZONE(zone)
	return (GetRealZoneText() == zone)
end

local function USABLE_FLYING()
	return (GetWintergraspWaitTime()or GetZoneText()~=L["Wintergrasp"])and IsFlyableArea()
end

local function USABLE_PROFESSION(prof, minimum)
	return (GetSkillLine(prof) and GetSkillLine(prof) >= minimum)
end

local function USABLE_ALWAYS()
	return true
end

local function USABLE_REAGENT(item)
	return (GetItemCount(item) > 0)
end

--====================
-- Lookup tables
--====================
local passengerMounts = {
	[55531] = 1, -- Mekgineer's Chopper
	[60424] = 1, -- Mechano-hog
	[61425] = 2, -- Traveler's Tundra Mammoth (A)
	[61447] = 2, -- Traveler's Tundra Mammoth (H)
	[61465] = 2, -- Grand Black War Mammoth (A)
	[61467] = 2, -- Grand Black War Mammoth (H)
	[61469] = 2, -- Grand Ice Mammoth (H)
	[61470] = 2, -- Grand Ice Mammoth (A)
	[75973] = 2, -- X-53 Touring Rocket
}

-- These mounts change stats periodically; thus, caching is ineffective. Manually refresh their data as necessary.
local variableMounts = {
	[58983] = {canBeLand = true, canBeFlight = false, MaxFlySpeed = nil}, -- Big Blizzard Bear
	[54729] = {canBeLand = false, canBeFlight = true, MaxFlySpeed = 280}, -- Winged Steed of the Ebon Blade
	[48025] = {canBeLand = true, canBeFlight = true, MaxFlySpeed = 280}, -- Headless Horseman's Mount
	[71342] = {canBeLand = true, canBeFlight = true, MaxFlySpeed = 310}, -- Big Love Rocket
	[75973] = {canBeLand = false, canBeFlight = true, MaxFlySpeed = 310}, -- X-53 Touring Rocket
	[72286] = {canBeLand = true, canBeFlight = true, MaxFlySpeed = 310}, -- Invincible
	[75614] = {canBeLand = true, canBeFlight = true, MaxFlySpeed = 310}, -- Celestial Steed
}

--====================
-- The guts
--====================
local function GenerateMountStats(t, spellID)
	local line = linecache[3]:GetText() -- WARNING. ASSUMES ALL MOUNT INFO LIES ON LINE 3.
	
	local mountType, passengers, speed, usable, varies
	
	passengers = passengerMounts[spellID] or 0
	if variableMounts[spellID] then
		local ridingSkill = GetSkillLine(L["Riding"])
		varies = true
		local canBeLand = variableMounts[spellID].canBeLand
		local canBeFlight = variableMounts[spellID].canBeFlight
		if (canBeLand) and (not canBeFlight) then
			speed = (ridingSkill >= 150 and 100) or 60
			mountType = MOUNTTYPE_LAND
			usable = USABLE_ALWAYS
		elseif (not canBeLand) and (canBeFlight) then
			if Has310Mount() then
				speed = variableMounts[spellID].MaxFlySpeed
			else
				speed = (ridingSkill == 300 and 280) or 150
			end
			mountType = MOUNTTYPE_FLYING
			usable = USABLE_FLYING
		elseif (canBeLand) and (canBeFlight) then
			if ridingSkill >= 225 and IsFlyableArea() then
				if Has310Mount() then
					speed = variableMounts[spellID].MaxFlySpeed
				else
					speed = (ridingSkill == 300 and 280) or 150
				end
				mountType = MOUNTTYPE_FLYING
				usable = USABLE_FLYING
			else
				speed = (ridingSkill >= 150 and 100) or 60
				mountType = MOUNTTYPE_LAND
				usable = USABLE_ALWAYS
			end
		end
	else
		varies = false
		if strfind(line, L["Outland"]) and strfind(line, L["Northrend"]) then
			mountType = MOUNTTYPE_FLYING
			speed = strfind(line, LEGENDARY) and 310 or (strfind(line, EPIC) and 280 or 150)
			usable = USABLE_FLYING
		elseif strfind(line, L["This mount can't move very quickly on land, but she's a great swimmer."]) then
			mountType = MOUNTTYPE_AQUATIC
			speed = 0
			usable = USABLE_ALWAYS
		else
			mountType = MOUNTTYPE_LAND
			speed = strfind(line, EPIC) and 100 or 60
			usable = USABLE_ALWAYS
		end
	end

	-- assert(type(speed) == "number", format("Invalid data for mount %d",spellID))
	-- assert(type(mountType) == "number", format("Invalid data for mount %d",spellID))
	-- assert(type(passengers) == "number", format("Invalid data for mount %d",spellID))
	-- assert(type(usable) == "function", format("Invalid data for mount %d",spellID))
	-- assert(type(varies) == "boolean", format("Invalid data for mount %d",spellID))
	if t then
		t[spellID] = {speed, mountType, passengers, usable, varies}
		return t[spellID]
	else
		return {speed, mountType, passengers, usable, varies}
	end	
end

-- Cache mount stats in a table
-- Some mounts possess oddities that would make generating their attributes dynamically senseless, or more effort than it's worth.
-- Instead, they're hardcoded into the cache.
local MountsCache = setmetatable( {
	-- AQ40 mounts
	[25953] = {100, MOUNTTYPE_LAND, 0, function() return USABLE_ZONE(L["Ahn'Qiraj"]) end}, -- Blue Qiraji Battle Tank
	[26054] = {100, MOUNTTYPE_LAND, 0, function() return USABLE_ZONE(L["Ahn'Qiraj"]) end}, -- Red Qiraji Battle Tank
	[26055] = {100, MOUNTTYPE_LAND, 0, function() return USABLE_ZONE(L["Ahn'Qiraj"]) end}, -- Yellow Qiraji Battle Tank
	[26056] = {100, MOUNTTYPE_LAND, 0, function() return USABLE_ZONE(L["Ahn'Qiraj"]) end}, -- Green Qiraji Battle Tank
	-- Tooltip inconsistency in enUS
	[60024] = {310, MOUNTTYPE_FLYING, 0, USABLE_FLYING}, -- Violet Proto-Drake
	-- Tooltip inconsistency in frFR
	[32242] = {280, MOUNTTYPE_FLYING, 0, USABLE_FLYING}, -- Swift Blue Gryphon
	[32289] = {280, MOUNTTYPE_FLYING, 0, USABLE_FLYING}, -- Swift Red Gryphon
	[32290] = {280, MOUNTTYPE_FLYING, 0, USABLE_FLYING}, -- Swift Green Gryphon
	[32292] = {280, MOUNTTYPE_FLYING, 0, USABLE_FLYING}, -- Swift Purple Gryphon
	-- Tailoring mounts
	[61451] = {150, MOUNTTYPE_FLYING, 0, function() return USABLE_FLYING() and USABLE_PROFESSION(L["Tailoring"], 410) end}, -- Flying Carpet
	[61309] = {280, MOUNTTYPE_FLYING, 0, function() return USABLE_FLYING() and USABLE_PROFESSION(L["Tailoring"], 425) end}, -- Magnificent Flying Carpet
	[75596] = {280, MOUNTTYPE_FLYING, 0, function() return USABLE_FLYING() and USABLE_PROFESSION(L["Tailoring"], 425) end}, -- Frosty Flying Carpet
	-- Engineering mounts
	[44151] = {280, MOUNTTYPE_FLYING, 0, function() return USABLE_FLYING() and USABLE_PROFESSION(L["Engineering"], 375) end}, -- Turbo-Charged Flying Machine
	[44153] = {150, MOUNTTYPE_FLYING, 0, function() return USABLE_FLYING() and USABLE_PROFESSION(L["Engineering"], 350) end}, -- Flying Machine
},{
	__index = GenerateMountStats
})

local critter_usable = setmetatable({}, {__index = function(t, i)
	for i=1,tip:NumLines() do
		local item = strfind(linecache[i]:GetText(), L["Reagents:"])
		-- NOTE: Cheating here. The only pets that require reagents as of 3.3 require Snowballs.
		-- Figuring out other items would require stripping escapes from the string.
		if item then
			t[i] = function() return USABLE_REAGENT(17202) end
			return t[i]
		end
	end
	t[i] = USABLE_ALWAYS
	return USABLE_ALWAYS
end})

addon.GetExtraCompanionData = function(companionType, spellID)
	assert(companionType=="CRITTER" or companionType=="MOUNT")
	SetTooltip(spellID)

	if companionType == "CRITTER" then
		return critter_usable[spellID]
	elseif companionType == "MOUNT" then
		if variableMounts[spellID] then
			return GenerateMountStats(nil, spellID)
		else
			return MountsCache[spellID]
		end
	end
end

